"""
Graphical user interface with the tkinter package to visualize data
readout from a Milli Gauss meter
"""


import tkinter, time
import tkinter.messagebox
import numpy as np
import pylab as p
# import seaborn as sb

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

from . import magnetometer as magneto 
from . import __version__, get_logger, create_timestamped_file

class GaussMeterGraphical(object):
    """
    A TKinter widget to visualize Gauss meter data
    """

    def __init__(self, master, device="/dev/ttyUSB0", interval=2,\
                 maxpoints=200, loglevel=20,
                 publish=True, port=9876):
        """
        Initialize the application window

        Args:
            master (tkinter.Tk): A tkinter main application

        Keyword Args:
            interval (int): Update the plot every interval seconds
            maxpoints (int): Max number of points visible in the plot
            loglevel (int): 10: debug, 20: info, 30: warnlng....

        """

        # Create container and menus
        self.master = master
        self.logger = get_logger(loglevel)
        self.frame = tkinter.Frame(self.master)
        top = self.master.winfo_toplevel()
        self.menu_bar = tkinter.Menu(top)
        top['menu'] = self.menu_bar

        self.sub_menu_help = tkinter.Menu(self.menu_bar)
        self.sub_menu_plot = tkinter.Menu(self.menu_bar)
        self.menu_bar.add_cascade(label='Plot', menu=self.sub_menu_plot)
        self.menu_bar.add_cascade(label='Help', menu=self.sub_menu_help)
        self.sub_menu_help.add_command(label='About', command=self._about_handler)
        self.sub_menu_plot.add_command(label="Reset", command=self.init_plot)
        self.sub_menu_plot.add_command(label="Log to file", command=self.init_datafile)

        # physics quantities
        self.meter = magneto.GaussMeterGU3001D(device=device,\
                                               loglevel=loglevel,\
                                               publish=publish,\
                                               port=port)
        self.start_time = time.monotonic()
        self.interval = interval
        self.maxpoints = maxpoints

        # writing results to file
        self.datafile_active = False
        self.datafilename = None
        self.datafile = None

        # plot
        fig = Figure()
        self.ax = fig.gca()
        self.canvas = FigureCanvasTkAgg(fig, master=self.master)
        self.canvas.show()
        self.canvas.get_tk_widget().pack(side='top', fill='both', expand=1)
        self.frame.pack()
        self.init_plot()
        self.update()

    def init_plot(self):
        """
        Initialize the plot
        """

        unit = self.meter.unit
        self.ax.set_xlabel("measurement time [s]")
        self.ax.set_ylabel("Magnetic field [{}]".format(unit))
        self.line, = self.ax.plot(range(0), color="blue", lw=3)

    @staticmethod
    def _about_handler():
        """
        Action performed if "about" menu item is clicked

        """
        tkinter.messagebox.showinfo("About", "Version: {}".format(__version__))

    def update(self):
        """
        Update the plot with recent magnetometer data
        """

        secs, fields = self.line.get_data()
        try:
            field = self.meter.measure()
        except Exception as e:
            self.logger.warning("Can not acquire data!")
            return
        sec = time.monotonic() - self.start_time

        # make sure data in the plot is "falling over"
        # so that it does not get too crammed
        index = 0
        if len(secs) >= self.maxpoints:
            self.logger.debug("Restricting line to {} points".format(self.maxpoints))
            index = 1

        secs = np.append(secs[index:], sec)
        fields = np.append(fields[index:], field)

        datamin = min(fields)
        datamax = max(fields)
        xmin = min(secs)
        xmax = max(secs)

        # avoid matplotlib warning
        if abs(datamin - datamax) < 1:
            datamin -= 1
            datamax += 1

        if abs(xmax - xmin) < 1:
            xmin -= 1
            xmax += 1

        # write to the datafile if desired
        if self.datafile_active:
            self.datafile.write("{:4.2f} {:4.2f}\n".format(sec, field))
        self.ax.set_xlim(xmin=xmin, xmax=xmax)
        self.ax.set_ylim(ymin=datamin, ymax=datamax)
        self.line.set_ydata(fields)
        self.line.set_xdata(secs)
        self.canvas.draw()
        self.master.after(self.interval, self.update)

    def init_datafile(self):
        """
        Write measurement results to a logfile
        """
        if self.datafile_active:
            self.datafile.close()

        self.datafilename = create_timestamped_file("GAUSSMETER_GU3001D_", file_ending=".dat")
        self.logger.info("Writing to file {}".format(self.datafilename))
        tkinter.messagebox.showinfo("Writing to a file!", "Writing data to file {}".format(self.datafilename))
        self.datafile = open(self.datafilename, "w")
        self.datafile.write("# seconds {}\n".format(self.meter.unit))
        self.datafile_active = True

    def __del__(self):
        """
        Close open files
        """
        if self.datafile_active:
            self.datafile.close()





"""
Read out some gaussmeter via USB connection
"""

import serial as s
import numpy as np
import time
import pylab as p
import seaborn as sb
import argparse

try:
    import zmq
except ImportError:
    print("Can not import zero MQ")

from . import get_logger

# example word... \r\x0241B20200000052\

# Connection settings:
# Baud rate 9600 
# Parity  No parity
# Data bit no. 8 Data bits
# Stop bit  1 Stop 
#

def save_execute(func):
    """
    Calls func with args, kwargs and 
    returns nan when func raises a value error
    
    FIXME: treatment of the errors!
    """
    def wrap_f(*args,**kwargs):
        try:
            return func(*args,**kwargs)
        except ValueError:
            return np.nan
    return wrap_f


#@save_execute
def decode_fields(data):
    """
    Give meaning to the fields of a 16 byte
    word returned by the meter. Words are 
    separated by \r

    Args:
        data (bytes): A single word of length 16 
                      
    """
    #### from the manual...
    # http://www.sunwe.com.tw/lutron/GU-3001eop.pdf
    # The 16 digits data stream will be displayed in the
    # following  format :
    # D15 D14 D13 D12 D11 D10 D9 D8 D7 D6 D5 D4 D3 D2 D1 D0
    # Each digit indicates the following status :
    # D15  Start Word = 02
    # D14  4
    # D13  When send the upper display data = 1
    #      When send the lower display data = 2
    # D12 & Annunciator for Display
    # D11   mG = B3
    #       uT = B2
    # D10 Polarity 
    #       0 = Positive
    #       1 = Negative
    # D9  Decimal Point(DP), position from right to the
    #       left, 0 = No DP, 1= 1 DP, 2 = 2 DP, 3 = 3 DP
    # D8 to D1  Display reading, D8 = MSD, D1 = LSD
    #           For example : 
    #           If the display reading is 1234, then D8 to
    #           D1 is : 00001234
    #          D0 End Word = 0D

    if not len(data) == 15:
        return np.nan
    polarity = data[5:6].decode()
    decimal_point = data[6:7].decode()
    mg_data = data[7:15].decode()
    polarity = int(polarity)
    decimal_point = int(decimal_point)
    if polarity:
        polarity = -1
    else:
        polarity = 1
    if decimal_point:
        index = -1*int(decimal_point)
        mg_data = mg_data[:index] + "." + mg_data[index:]
    mg_data = polarity*float(mg_data)
    return mg_data


def get_unit(data):
    """
    Get the unit from a magnetometer word
    """
    if not len(data) == 15:
        raise ValueError("Data corrupted!")
    unit = data[3:5].decode()    
    if unit == "B3":
        unit = "mG"
    elif unit == "B2":
        unit = "uT"
    else:
        raise ValueError("Unit not understood {}".format(unit))
    return unit 

def decode_meter_output(data):
    """
    Decode the bytestring with 
    hex numbers to the field 
    information
    
    Args:
        data (bytes): raw output from serial port

    """
    data = data.split(b"\r")
    data = [decode_fields(word) for word in data]
    data = np.array(data)
    data = data[np.isfinite(data)]
    return data


class GaussMeterGU3001D(object):
    """
    The named instrument
    """

    def __init__(self, device="/dev/ttyUSB0", loglevel=20,\
                 publish=False, port=9876):
        """
        Constructor needs read and write access to 
        the port

        Keyword Args:
            port (str): The virtual serial connection on a UNIX system
            loglevel (int): 10: debug, 20: info, 30: warnlng....
            publish (bool): publish data on port
            port (int): use this port if publish = True
        """
        self.meter = s.Serial(device) # default settings are ok
        self.logger = get_logger(loglevel)
        self.logger.debug("Meter initialized")
        self.publish = publish
        self.port = port
        self._socket = None

    def _setup_port(self):
        """
        Setup the port for publishing

        Returns:
            None
        """
        context = zmq.Context()
        self._socket = context.socket(zmq.PUB)
        self._socket.bind("tcp://*:%s" % int(self.port))
        return

    @property
    def unit(self):
        """
        Figure out the units of the meter
        """
        time.sleep(1)
        some_data = self.meter.read_all()
        some_data = some_data.split(b"\r")[0]
        unit = None
        while unit is None:
            try:
                unit = get_unit(some_data)
            except ValueError:
                self.logger.debug("Can not get unit from {}, trying again...".format(some_data))
                time.sleep(2)
                some_data = self.meter.read_all()
                some_data = some_data.split(b"\r")[0]
        self.logger.info("All data will be in {}".format(unit))
        return unit    

    def measure(self, measurement_time=1):
        """
        Measure a single point

        Keyword Args:
            measurement_time (int): average the values over the measurement time

        """
        time.sleep(measurement_time) # give the meter time to acquire some data
        data = self.meter.read_all()
        unit = get_unit(data.split(b"\r")[0])
        field = decode_meter_output(data)
        field = field.mean()
        if self.publish and (self._socket is None):
            self._setup_port()
        if self.publish:
            self._socket.send_string("{} {}; {}".format("GU3001D", field, self.unit))
        return field         

    def measure_continously(self, npoints, interval):
        """
        Make a measurement with npoints each interval seconds
    
        Args:
            npoints (int): number of measurement points
            interval (int): make a measurement each interval seconds
            
        Keyword Args:
            silent (bool): Suppress output

        """

        for n in range(npoints):
            field = self.measure(measurement_time=interval)
            yield n*interval, field        

 



"""
Read out a "milligauss-meter" GU-3001D
"""

import logging
import sys

from matplotlib import get_configdir as mpl_configdir
from os.path import split as _split
from os.path import join as _join
from os.path import exists as _exists
from shutil import copy as _copy
from datetime import datetime as _datetime

__version__ = "0.0.8"
__all__ = ["magnetometer","gui"]

_appdir = _split(__file__)[0]

LOGFORMAT = '[%(asctime)s] %(levelname)s: %(module)s(%(lineno)d):   %(message)s'
STYLE_BASEFILE_STD = _join(_appdir,"goldschmidt.mplstyle")

logging.captureWarnings(True)


def create_timestamped_file(filename, file_ending=".log"):
    """
    Return a timestamped filename (with the date in it) which is not yet
    present on the system
    Args:
        filename (str): Name of the file (full path) (without ending)

    Keyword Args:
        file_ending (str): The fileending of the new file

    Returns:
        str
    """
    today = _datetime.now()
    today = today.strftime("%Y_%m_%d_%H_%M")
    if filename.endswith(file_ending):
        filename.replace(file_ending, today + file_ending)
    else:
        filename += (today + file_ending)
    filecount = 1
    while _exists(filename):
        filename = filename.replace("." + str(filecount - 1), "")
        filename = filename + "." + str(filecount)
        filecount += 1
        if filecount >= 60:
            raise SystemError("More than 1 file per second, this is insane.. aborting")
    return filename

def get_logger(loglevel, logfile=None):
    """
    A root logger with a formatted output logging to stdout and a file

    Args:
        loglevel (int): 10,20,30,... the higher the less logging

    Keyword Args:
        logfile (str): write logging to this file as well as stdout

    """

    def exception_handler(exctype, value, tb):
        logger.critical("Uncaught exception", exc_info=(exctype, value, tb))

    logger = logging.getLogger()
    logger.setLevel(loglevel)
    ch = logging.StreamHandler()
    ch.setLevel(loglevel)
    formatter = logging.Formatter(LOGFORMAT)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    if logfile is not None:
        logfile = create_timestamped_file(logfile, file_ending=".log")
        fh = logging.FileHandler(logfile)
        fh.setFormatter(formatter)
        fh.setLevel(loglevel)
        logger.addHandler(fh)

    sys.excepthook = exception_handler
    logger.propagate = False
    return logger


def install_styles(style_default=STYLE_BASEFILE_STD):
    """
    Sets up style files

    Keyword Args:
        style_default (str): location of style file to use by defautl
   """

    mpl_styledir = _join(mpl_configdir(),"stylelib")
    assert _exists(style_default), "STYLEFILE {} missing... indicates a problem with some paths or corrupt packege.\
                                    Check source code location".format(style_default)
    _copy(style_default,mpl_styledir)

install_styles()



from setuptools import setup

from goldschmidt import __version__

setup(name='goldschmidt',
      version=__version__,
      description='Readout a GU-3001D Milli-Gauss meter via USB/serial port on a linux system',
      long_description='Provides commandline readout and gui for the named instrument. For the name see https://en.wikipedia.org/wiki/Carl_Wolfgang_Benjamin_Goldschmidt who was an astronomer, mathematician and physicist and assistant to Gauss',
      author='Achim Stoessl',
      author_email="achim.stoessl@gmail.com",
      url='https://github.com/achim1/goldschmidt',
      #download_url="pip install pyosci",
      install_requires=['numpy>=1.11.0',
                        'matplotlib>=1.5.0'],
      license="GPL",
      platforms=["Ubuntu 14.04","Ubuntu 16.04"],
      classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Physics"
              ],
      keywords=["magnetometer", "gauss",\
                "GU-3001D", "gaussmeter", "Milli-Gauss Meter",\
                "readout", "physics", "engineering", "serial", "RS-232", "USB"],
      packages=['goldschmidt'],
      scripts=["bin/goldschmidt"],
      package_data={'goldschmidt': ['goldschmidt.mplstyle']}
      )
